# This file should never be used for anything. It exists, so that VSCode
# recognizes this folder as a Python package. This in turn allows components and
# pages to reference each other, thus enabling refactoring.

from .sample_component import *
